//
//	Copyright  Nemetschek Vectorworks, Inc.
//	Use of this file is governed by the Nemetschek Vectorworks SDK License Agreement
//	http://developer.vectorworks.net/index.php?title=Vectorworks_SDK_License
//

#include "StdHeaders.h"
//#include "VWRecordObj.h"

using namespace VWFC::Tools;
using namespace VWFC::VWObjects;

VWRecordObj::VWRecordObj(MCObjectHandle hRecordHandle)
{
	fhObject		= hRecordHandle;

	VWFC_ASSERT( VWRecordObj::IsRecordObject( fhObject ) );
	if ( ! VWRecordObj::IsRecordObject( fhObject ) )
		THROW_VWFC_EXCEPTION( VWBadObjectTypeException, 0, "bad handle type when creating" );
}

// bUseExistingRecord	- true then use the record in hObject of format hFormat if it exist - otherwise create record
//						- false then always crete record object from the hFormat inside hObject
VWRecordObj::VWRecordObj(MCObjectHandle hFormat, MCObjectHandle hObject, bool bUseExistingRecord)
{
	if ( hFormat == NULL ) {
		THROW_VWFC_EXCEPTION( VWBadObjectTypeException, 0, "bad format handle type when creating" );
	}
	else {
		MCObjectHandle		hRecord	= NULL;
		if ( bUseExistingRecord ) {
			TFormatHandler		fh( hFormat );
			hRecord						= fh.AttachRecordToObject( hObject ); 
		}
		else {
			MCObjectHandle	hDefaultRecord	= ::GS_FindAuxObject( gCBP, hFormat, kRecordNode );
			ASSERTN( kVStanev, hDefaultRecord );
			hRecord							= ::GS_DuplicateObject( gCBP, hDefaultRecord );
			::GS_AttachAuxObject( gCBP, hRecord, hObject );

			// find the last object in the AUX list
			MCObjectHandle	hLastObj	= NULL;
			MCObjectHandle	hNextObj	= hRecord;
			while( hNextObj ) {
				MCObjectHandle	hNextAux	= ::GS_NextObject( gCBP, hNextObj );
				hNextObj					= ::GS_NextAuxObject( gCBP, hNextAux, kRecordNode );
				if ( hNextObj )
					hLastObj	= hNextObj;
			}

			// attach this record at the end of the AUX list
			if ( hLastObj ) {
				::GS_InsertObjectAfter( gCBP, hRecord, hLastObj );
			}
		}

		fhObject					= hRecord;

		VWFC_ASSERT( VWRecordObj::IsRecordObject( fhObject ) );
		if ( ! VWRecordObj::IsRecordObject( fhObject ) )
			THROW_VWFC_EXCEPTION( VWBadObjectTypeException, 0, "bad handle type when creating" );
	}
}

VWRecordObj::VWRecordObj(const VWRecordObj& src)
{
	fhObject	= src.fhObject;
}

VWRecordObj::~VWRecordObj()
{
}

VWRecordObj& VWRecordObj::operator=(const VWRecordObj& src)
{
	fhObject	= src.fhObject;
	return *this;
}

TXString VWRecordObj::GetRecordName() const
{
	VWFC_ASSERT( fhObject );
	TRecordHandler	rh( fhObject );
	
	Str63 nameOfFormat;
	rh.GetNameOfFormat( nameOfFormat );

	return nameOfFormat;
}

TXString VWRecordObj::GetRecordLocalizedName() const
{
	TXString	localizedName;

	VWFC_ASSERT( fhObject );
	TRecordHandler	rh( fhObject );
	if ( rh.IsParametric() ) {
		TVariableBlock	varName;
		gSDK->GetObjectVariable( fhObject, ovParametricLocalizedName, varName );
		PStr255			theString;
		if ( varName.GetStr255( theString ) ) {
			localizedName	= theString;
		}
		else {
			localizedName	= this->GetRecordName();
		}
	}
	else {
		localizedName	= this->GetRecordName();
	}

	return localizedName;
}

bool VWRecordObj::IsParametric() const
{
	VWFC_ASSERT( fhObject );
	TRecordHandler	rh( fhObject );
	return rh.IsParametric() ? true : false;
}

void VWRecordObj::SetParametric(bool bIsParametric)
{
	VWFC_ASSERT( fhObject );
	TRecordHandler	rh( fhObject );
	rh.SetIsParametric( bIsParametric );
}

size_t VWRecordObj::GetParamsCount() const
{
	VWFC_ASSERT( fhObject );
	TRecordHandler	rh( fhObject );
	return (size_t) rh.GetNumOfFields();
}

TXString VWRecordObj::GetParamName(size_t paramIndex) const
{
	VWFC_ASSERT( fhObject );
	TRecordHandler	rh( fhObject );

	
	TXString	nameOfField;
	// NOTE! 'GetNameOfField' return 1 based value
	rh.GetNameOfField( DemoteTo<short>( kVStanev, paramIndex+1 ), nameOfField );
	return nameOfField;
}

size_t VWRecordObj::GetParamIndex(const TXString& paramName) const
{
	VWFC_ASSERT( fhObject );
	TRecordHandler	rh( fhObject );

	// NOTE! 'GetFieldIndex' return 1 based value
	int		index	= rh.GetFieldIndex( paramName );
	if ( index == 0 )
		return (size_t) -1;	// not found

	return (size_t) (index - 1);
}

EFieldStyle VWRecordObj::GetParamStyle(size_t paramIndex) const
{
	VWFC_ASSERT( fhObject );
	TRecordHandler	rh( fhObject );

	// NOTE! 'GetFieldStyle' return 1 based value
	return (EFieldStyle) rh.GetFieldStyle( DemoteTo<short>( kVStanev, paramIndex+1 ) );
}

TXString VWRecordObj::GetParamValue(size_t paramIndex) const
{
	VWFC_ASSERT( fhObject );
	TRecordHandler	rh( fhObject );

	// NOTE! itemIndex is expected from the SDK as 1 based value
	short			itemIndex	= DemoteTo<short>( kVStanev, paramIndex+1 );
	EFieldStyle		style		= (EFieldStyle) rh.GetFieldStyle( itemIndex );
	TRecordItem valueItem( style );
	Boolean			ok			= rh.GetFieldObject( itemIndex, valueItem );
	VWFC_ASSERT( ok );

	TXString	value;
	if ( style == kFieldText ) {
		char*		string		= NULL;
		size_t		len			= 0;

		if ( ok ) {
			valueItem.GetFieldValue( string, len );
			value	= TXString( string, len );
		}
	}
	else {
		TXString		theString;
		if ( valueItem.GetFieldValueAsString( theString ) ) {
			value	= theString;
		}
	}

	return value;
}

void VWRecordObj::SetParamValue(size_t paramIndex, const TXString& value)
{
	VWFC_ASSERT( fhObject );
	TRecordHandler	rh( fhObject );

	// NOTE! itemIndex is expected from the SDK as 1 based value
	short			itemIndex	= DemoteTo<short>( kVStanev, paramIndex+1 );
	EFieldStyle		style		= (EFieldStyle) rh.GetFieldStyle( itemIndex );
	TRecordItem		recItem( style );
	if ( VWFC_VERIFY( rh.GetFieldObject( itemIndex, recItem ) ) ) {
		if ( style == kFieldText ) {
			const char* szText		= value;
			size_t		textLen		= value.GetLength();

			if ( VWFC_VERIFY( recItem.SetFieldValue( szText, textLen ) ) ) {
				VWFC_VERIFY( rh.SetFieldObject( itemIndex, recItem ) );
			}
		}
		else {
			
			if ( VWFC_VERIFY( recItem.SetFieldValueAsString( value ) ) ) {
				VWFC_VERIFY( rh.SetFieldObject( itemIndex, recItem ) );
			}
		}
	}
}


long VWRecordObj::GetParamLong(size_t paramIndex) const
{
	VWFC_ASSERT( fhObject );
	TRecordHandler	rh( fhObject );

	// NOTE! itemIndex is expected from the SDK as 1 based value
	short		itemIndex	= DemoteTo<short>( kVStanev, paramIndex+1 );
	TRecordItem valueItem( rh.GetFieldStyle( itemIndex ) );
	Boolean		ok			= rh.GetFieldObject( itemIndex, valueItem );
	VWFC_ASSERT( ok );

	long		value	= 0;
	if ( ok ) {
		valueItem.GetFieldValue( value );
	}

	return value;
}

void VWRecordObj::SetParamLong(size_t paramIndex, long value)
{
	VWFC_ASSERT( fhObject );
	TRecordHandler	rh( fhObject );

	// NOTE! itemIndex is expected from the SDK as 1 based value
	short			itemIndex	= DemoteTo<short>( kVStanev, paramIndex+1 );
	TRecordItem		recItem( rh.GetFieldStyle( itemIndex ) );
	if ( VWFC_VERIFY( rh.GetFieldObject( itemIndex, recItem ) ) ) {
		if ( VWFC_VERIFY( recItem.SetFieldValue( value ) ) ) {
			VWFC_VERIFY( rh.SetFieldObject( itemIndex, recItem ) );
		}
	}
}

bool VWRecordObj::GetParamBool(size_t paramIndex) const
{
	VWFC_ASSERT( fhObject );
	TRecordHandler	rh( fhObject );

	// NOTE! itemIndex is expected from the SDK as 1 based value
	short		itemIndex	= DemoteTo<short>( kVStanev, paramIndex+1 );
	TRecordItem valueItem( rh.GetFieldStyle( itemIndex ) );
	Boolean		ok			= rh.GetFieldObject( itemIndex, valueItem );
	VWFC_ASSERT( ok );

	Boolean		value	= 0;
	if ( ok ) {
		valueItem.GetFieldValue( value );
	}

	return value ? true : false;
}

void VWRecordObj::SetParamBool(size_t paramIndex, bool value)
{
	VWFC_ASSERT( fhObject );
	TRecordHandler	rh( fhObject );

	// NOTE! itemIndex is expected from the SDK as 1 based value
	short			itemIndex	= DemoteTo<short>( kVStanev, paramIndex+1 );
	TRecordItem		recItem( rh.GetFieldStyle( itemIndex ) );
	if ( VWFC_VERIFY( rh.GetFieldObject( itemIndex, recItem ) ) ) {
		Boolean		boolValue	= value;
		if ( VWFC_VERIFY( recItem.SetFieldValue( boolValue ) ) ) {
			VWFC_VERIFY( rh.SetFieldObject( itemIndex, recItem ) );
		}
	}
}

double VWRecordObj::GetParamReal(size_t paramIndex) const
{
	VWFC_ASSERT( fhObject );
	TRecordHandler	rh( fhObject );

	// NOTE! itemIndex is expected from the SDK as 1 based value
	short		itemIndex	= DemoteTo<short>( kVStanev, paramIndex+1 );
	TRecordItem valueItem( rh.GetFieldStyle( itemIndex ) );
	Boolean		ok			= rh.GetFieldObject( itemIndex, valueItem );
	VWFC_ASSERT( ok );

	double		value	= 0;
	if ( ok ) {
		valueItem.GetFieldValue( value );
	}

	return value;
}

void VWRecordObj::SetParamReal(size_t paramIndex, double value)
{
	VWFC_ASSERT( fhObject );
	TRecordHandler	rh( fhObject );

	// NOTE! itemIndex is expected from the SDK as 1 based value
	short			itemIndex	= DemoteTo<short>( kVStanev, paramIndex+1 );
	TRecordItem		recItem( rh.GetFieldStyle( itemIndex ) );
	if ( VWFC_VERIFY( rh.GetFieldObject( itemIndex, recItem ) ) ) {
		if ( VWFC_VERIFY( recItem.SetFieldValue( value ) ) ) {
			VWFC_VERIFY( rh.SetFieldObject( itemIndex, recItem ) );
		}
	}
}

TXString VWRecordObj::GetParamString(size_t paramIndex) const
{
	VWFC_ASSERT( fhObject );
	TRecordHandler	rh( fhObject );

	// NOTE! itemIndex is expected from the SDK as 1 based value
	short		itemIndex	= DemoteTo<short>( kVStanev, paramIndex+1 );
	TRecordItem valueItem( rh.GetFieldStyle( itemIndex ) );
	Boolean		ok		= rh.GetFieldObject( itemIndex, valueItem );
	VWFC_ASSERT( ok );

	TXString	value;
	char*		string		= NULL;
	size_t		len			= 0;

	if ( ok ) {
		valueItem.GetFieldValue( string, len );
		value	= TXString( string, len );
	}

	return value;
}

void VWRecordObj::SetParamString(size_t paramIndex, const TXString& value)
{
	VWFC_ASSERT( fhObject );
	TRecordHandler	rh( fhObject );

	// NOTE! itemIndex is expected from the SDK as 1 based value
	short			itemIndex	= DemoteTo<short>( kVStanev, paramIndex+1 );
	TRecordItem		recItem( rh.GetFieldStyle( itemIndex ) );
	if ( VWFC_VERIFY( rh.GetFieldObject( itemIndex, recItem ) ) ) {
		const char* szText		= value;
		size_t		textLen		= value.GetLength();

		if ( VWFC_VERIFY( recItem.SetFieldValue( szText, textLen ) ) ) {
			VWFC_VERIFY( rh.SetFieldObject( itemIndex, recItem ) );
		}
	}
}


size_t VWRecordObj::GetParamIndexByUnivName(const TXString& univParamName) const
{
	size_t	paramIndex	= size_t(-1);

	// this little hach but we want to be able to change the this
	// this is needed for performance optimization
	// so we would add the parameters only when they are needed
	// not for every parameter in the contrustor
	VWRecordObj*	pThis = (VWRecordObj*)( this );

	TXString	macUnivParamName	= univParamName;
	macUnivParamName.AnsiToMac();
	TUnviNameKeyMap::iterator	it = pThis->fmapUnivNameKey.find( macUnivParamName );
	if ( it == pThis->fmapUnivNameKey.end() ) {
		TRecordHandler rh( fhObject );
		// NOTE! 'GetFieldIndex' return 1 based value
		int		index	= rh.GetFieldIndex( macUnivParamName );
		if ( index != 0 ) {
			paramIndex	= (size_t) (index - 1);
		}

		pThis->fmapUnivNameKey.insert( TUnviNameKeyMap::value_type( macUnivParamName, paramIndex ) );
	}
	else {
		paramIndex	= it->second;
	}

	return paramIndex;
}

EFieldStyle VWRecordObj::GetParamStyle(const TXString& univParamName) const
{
	return this->GetParamStyle( this->GetParamIndexByUnivName( univParamName ) );
}

TXString VWRecordObj::GetParamValue(const TXString& univParamName) const
{
	return this->GetParamValue( this->GetParamIndexByUnivName( univParamName ) );
}

void VWRecordObj::SetParamValue(const TXString& univParamName, const TXString& value)
{
	this->SetParamValue( this->GetParamIndexByUnivName( univParamName ), value );
}

long VWRecordObj::GetParamLong(const TXString& univParamName) const
{
	return this->GetParamLong( this->GetParamIndexByUnivName( univParamName ) );
}

void VWRecordObj::SetParamLong(const TXString& univParamName, long value)
{
	this->SetParamLong( this->GetParamIndexByUnivName( univParamName ), value );
}

bool VWRecordObj::GetParamBool(const TXString& univParamName) const
{
	return this->GetParamBool( this->GetParamIndexByUnivName( univParamName ) );
}

void VWRecordObj::SetParamBool(const TXString& univParamName, bool value)
{
	this->SetParamBool( this->GetParamIndexByUnivName( univParamName ), value );
}

double VWRecordObj::GetParamReal(const TXString& univParamName) const
{
	return this->GetParamReal( this->GetParamIndexByUnivName( univParamName ) );
}

void VWRecordObj::SetParamReal(const TXString& univParamName, double value)
{
	this->SetParamReal( this->GetParamIndexByUnivName( univParamName ), value );
}

TXString VWRecordObj::GetParamString(const TXString& univParamName) const
{
	return this->GetParamString( this->GetParamIndexByUnivName( univParamName ) );
}

void VWRecordObj::SetParamString(const TXString& univParamName, const TXString& value)
{
	this->SetParamString( this->GetParamIndexByUnivName( univParamName ), value );
}

/*static*/ bool VWRecordObj::IsRecordObject(MCObjectHandle hObj)
{
	short	type	= gSDK->GetObjectTypeN( hObj );
	return type == kRecordNode;
}

/*static*/ bool VWRecordObj::IsRecordObject(MCObjectHandle hObj, const TXString& recordName)
{
	bool	result	= false;
	short	type	= gSDK->GetObjectTypeN( hObj );
	if ( type == kRecordNode ) {
		VWFC_ASSERT( hObj );
		TRecordHandler	rh( hObj );
		
		Str63 nameOfFormat;
		rh.GetNameOfFormat( nameOfFormat );

		result	= (recordName.CompareNoCase( (StringPtr) nameOfFormat ) == 0);
	}

	return result;
}

/*static*/ MCObjectHandle VWRecordObj::GetRecordObject(MCObjectHandle hObj, const TXString& recordName)
{
	MCObjectHandle	hFoundRecord	= NULL;
	if ( hObj ) {
		VWObject			theObject( hObj );
		MCObjectHandle		hRecord		= theObject.FindAuxObject( kRecordNode );
		while ( hRecord && VWRecordObj::IsRecordObject( hRecord ) ) {
			TRecordHandler	rh( hRecord );
			
			Str63 nameOfFormat;
			rh.GetNameOfFormat( nameOfFormat );
	
			if ( recordName.CompareNoCase( (StringPtr) nameOfFormat ) == 0 ) {
				hFoundRecord	= hRecord;
				break;
			}

			hRecord		= theObject.FindAuxObject( kRecordNode, hRecord );
		}
	}

	return hFoundRecord;
}

